from nebula3.Config import Config
from nebula3.gclient.net import ConnectionPool


def test_networkx_engine_query_reader():
    import networkx as nx

    from ng_ai import NebulaReader
    from ng_ai.config import NebulaGraphConfig

    # read data with spark engine, query mode
    config_dict = {
        "graphd_hosts": "127.0.0.1:39669",
        "user": "root",
        "password": "nebula",
        "space": "basketballplayer",
    }
    config = NebulaGraphConfig(**config_dict)
    reader = NebulaReader(engine="nebula", config=config)
    reader.query(edges=["follow", "serve"], props=[["degree"], []])
    g = reader.read()
    assert isinstance(g.data, nx.MultiDiGraph)
    g.show(10)


def test_networkx_engine_algo():
    """
    Test networkx engine with all algorithms
    """

    from ng_ai import NebulaReader
    from ng_ai.config import NebulaGraphConfig

    # read data with spark engine, query mode
    config_dict = {
        "graphd_hosts": "127.0.0.1:39669",
        "user": "root",
        "password": "nebula",
        "space": "basketballplayer",
    }
    config = NebulaGraphConfig(**config_dict)
    reader = NebulaReader(engine="nebula", config=config)
    reader.query(edges=["follow", "serve"], props=[["degree"], []])
    g = reader.read()

    for algo_name in g.algo.get_all_algo():
        algo_func = getattr(g.algo, algo_name)
        print(f"Running {algo_name}...")
        result = algo_func()
        if hasattr(result, "__iter__"):
            for x in result:
                print(x)
        else:
            print(result)
        print("=" * 20)


def test_networkx_engine_writer():
    from ng_ai import NebulaReader, NebulaWriter
    from ng_ai.config import NebulaGraphConfig

    # read data with spark engine, query mode
    config_dict = {
        "graphd_hosts": "127.0.0.1:39669",
        "user": "root",
        "password": "nebula",
        "space": "basketballplayer",
    }
    config = NebulaGraphConfig(**config_dict)
    reader = NebulaReader(engine="nebula", config=config)
    reader.query(edges=["follow", "serve"], props=[["degree"], []])
    g = reader.read()

    graph_result = g.algo.louvain()

    writer = NebulaWriter(
        data=graph_result,
        sink="nebulagraph_vertex",
        config=config,
        engine="nebula",
    )

    # properties to write
    properties = ["cluster_id"]

    writer.set_options(
        tag="louvain",
        properties=properties,
        batch_size=256,
        write_mode="insert",
    )
    # write back to NebulaGraph
    writer.write()

    nebula_config = Config()
    connection_pool = ConnectionPool()
    connection_pool.init([("127.0.0.1", 39669)], nebula_config)

    with connection_pool.session_context("root", "nebula") as session:
        session.execute("USE basketballplayer")
        result = session.execute(
            "MATCH (v:louvain) RETURN id(v), v.louvain.cluster_id LIMIT 10;"
        )
        print(result)
    connection_pool.close()

    assert result.is_succeeded(), f"ERROR during query NebulaGraph: {result}"
    assert (
        not result.is_empty()
    ), f"louvain not written to NebulaGraph result: {result}"

    print(f"Label propagation result:\n{result}")



from ng_ai import NebulaReader, NebulaWriter
from ng_ai.config import NebulaGraphConfig

config_dict = {
    "graphd_hosts": "graphd:9669",
    "metad_hosts": "metad0:9559",
    "user": "root",
    "password": "nebula",
    "space": "basketballplayer",
}
config = NebulaGraphConfig(**config_dict)
reader = NebulaReader(engine="spark", config=config)
reader.scan(edge="follow", props="degree")
df = reader.read()

df_result = df.algo.label_propagation()

writer = NebulaWriter(
    data=df_result, sink="nebulagraph_vertex", config=config, engine="spark"
)

# map column louvain into property cluster_id
properties = {"lpa": "cluster_id"}

writer.set_options(
    tag="label_propagation",
    vid_field="_id",
    properties=properties,
    batch_size=256,
    write_mode="insert",
)
# write back to NebulaGraph
writer.write()

"""
/spark/bin/spark-submit \
    --jars /root/download/nebula-algo.jar \
    /root/run/writer.py
"""

# edge writer

df_result = df.algo.jaccard()

writer = NebulaWriter(
    data=df_result, sink="nebulagraph_edge", config=config, engine="spark"
)

# map column louvain into property cluster_id
properties = {"similarity": "similarity"}

writer.set_options(
    space="basketballplayer",
    edge_type="jaccard_similarity",
    src_id="srcId",
    dst_id="dstId",
    src_id_policy="",
    dst_id_policy="",
    properties=properties,
    batch_size=256,
    write_mode="insert",
)

# write back to NebulaGraph
writer.write()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.

from __future__ import annotations

from ng_ai.engines import NebulaEngine


class NebulaGraphObject:
    def __init__(self, engine: NebulaEngine, raw_graph, **kwargs):
        self.engine = engine
        # if engine is nebula, self._graph is a networkx graph object
        # if engine is spark, self._graph is a spark graph object
        self._graph = raw_graph
        # let's keep unified interface as NebulaDataFrameObject
        self.data = self._graph

    def get_engine(self):
        return self.engine

    @property
    def algo(self):
        if self.engine.type == "spark":
            print(
                "NebulaGraphObject.algo is not supported in spark engine, "
                "please use NebulaDataFrameObject.algo instead"
            )
            raise NotImplementedError
        if self.engine.type == "nebula":
            from ng_ai.nebula_algo import NebulaAlgorithm as NebulaAlgorithmImpl

            return NebulaAlgorithmImpl(self)

    def get_nx_graph(self):
        if self.engine.type == "nebula":
            return self._graph
        else:
            # for now the else case will be spark, to networkx is not supported
            raise Exception(
                "For NebulaGraphObject in spark engine,"
                "convert to networkx graph is not supported",
            )

    def stats(self, *keywords, **kwargs):
        if self.engine.type == "spark":
            raise NotImplementedError
        elif self.engine.type == "nebula":
            print(self.data)

    def draw(self, with_labels=True, **kwargs):
        if self.engine.type == "nebula":
            self.engine.nx.draw(self.data, with_labels=with_labels, **kwargs)
        else:
            raise NotImplementedError

    def show(self, limit=None):
        if self.engine.type == "spark":
            raise NotImplementedError
        elif self.engine.type == "nebula":
            if limit is None:
                limit = 0
            for u, v, key, attrs in self.data.edges(keys=True, data=True):
                print(f"{u} -> {v} ({key}): {attrs}")
                limit -= 1
                if limit == 0:
                    break


class NebulaDataFrameObject:
    def __init__(self, engine, data):
        """
        data: pd.DataFrame or spark.DataFrame
        """
        self.engine = engine
        self.data = data
        # if engine is nebula, self.data is a pandas dataframe
        # if engine is spark, self.data is a spark dataframe

    def get_engine(self):
        return self.engine

    @property
    def algo(self):
        if self.engine.type == "spark":
            from ng_ai.nebula_algo import NebulaAlgorithm as NebulaAlgorithmImpl

            return NebulaAlgorithmImpl(self)
        else:
            print(
                "NebulaDataFrameObject.algo is not supported in nebula engine, "
                "please use NebulaGraphObject.algo instead"
            )
            raise NotImplementedError

    def get_spark_df(self):
        if self.engine.type == "spark":
            return self.data
        else:
            # convert pandas dataframe to spark dataframe
            # not implemented now
            raise NotImplementedError

    def to_pandas_df(self):
        if self.engine.type == "nebula":
            return self.data
        else:
            # convert the spark df to a pandas data frame
            # we should carefully think about this use case
            # not implemented now
            raise NotImplementedError

    def to_networkx(self):
        # for now the else case will be spark, to networkx is not supported
        raise Exception(
            "For NebulaDataFrameObject in spark engine,"
            "convert to networkx graph is not supported",
        )

    def to_graphx(self):
        if self.engine.type == "spark":
            df = self.data  # noqa: F841
            # convert the df to a graphx graph, not implemented now
            raise NotImplementedError
        else:
            # for now the else case will be nebula, to graphx is not supported
            raise Exception(
                "For NebulaDataFrameObject in nebula engine,"
                "convert to graphx is not supported",
            )

    def to_graph(self):
        raise NotImplementedError

    def show(self, *keywords, **kwargs):
        if self.engine.type == "spark":
            self.data.show(*keywords, **kwargs)
        elif self.engine.type == "nebula":
            print(self.data)
        else:
            raise NotImplementedError



from ng_ai import NebulaGraphConfig, NebulaReader

config_dict = {
    "graphd_hosts": "graphd:9669",
    "metad_hosts": "metad0:9559",
    "user": "root",
    "password": "nebula",
    "space": "basketballplayer",
}
config = NebulaGraphConfig(**config_dict)
reader = NebulaReader(engine="spark", config=config)
reader.scan(edge="follow", props="degree")
df = reader.read()

pr_result = df.algo.pagerank(reset_prob=0.15, max_iter=10)

pr_result.show(5)

"""
/spark/bin/spark-submit \
    --jars /root/download/nebula-algo.jar \
    /root/run/algo.py
"""



import sys
from unittest.mock import MagicMock

import pytest
from pyspark.sql import SparkSession

from ng_ai import NebulaGraphConfig
from ng_ai.engines import SparkEngine
from ng_ai.nebula_algo import NebulaDataFrameAlgorithm
from ng_ai.nebula_data import NebulaDataFrameObject


# Pytest fixture to create a SparkSession instance
@pytest.fixture(scope="session")
def spark():
    spark = (
        SparkSession.builder.master("local[1]")
        .config("spark.port.maxRetries", "1000")
        .config("spark.sql.shuffle.partitions", "1")
        .getOrCreate()
    )
    return spark


# Pytest fixture to create a sample Spark DataFrame
@pytest.fixture(scope="session")
def spark_df(spark):
    """
    +---------+---------+-----+------+
    |   _srcId|   _dstId|_rank|degree|
    +---------+---------+-----+------+
    |player105|player100|    0|    70|
    |player105|player104|    0|    83|
    +---------+---------+-----+------+
    """
    data = [
        ("player105", "player100", 0, 70),
        ("player105", "player104", 0, 83),
    ]

    # spark engine can only be tested in python 3.9 or lower
    assert sys.version_info < (
        3,
        10,
    ), "spark engine can only be tested in python 3.9 or lower"
    df = spark.createDataFrame(data, ["_srcId", "_dstId", "_rank", "degree"])
    return df


def test_algo_dot_call_spark_engine(spark_df):
    config = NebulaGraphConfig()
    spark_engine = SparkEngine(config=config)
    nebula_df = NebulaDataFrameObject(engine=spark_engine, data=spark_df)

    mock_pagerank = MagicMock()
    NebulaDataFrameAlgorithm.pagerank = mock_pagerank

    nebula_df.algo.pagerank(reset_prob=0.15, max_iter=10)

    mock_pagerank.assert_called_once_with(reset_prob=0.15, max_iter=10)



import subprocess

from nebula3.Config import Config
from nebula3.gclient.net import ConnectionPool


def test_scan_reader_spark_engine():
    """
    Just call:
    /spark/bin/spark-submit \
        --jars /root/download/nebula-algo.jar \
        /root/run/scan_reader.py
    in container: master_spark
    assert the return value is 0
    """

    result = subprocess.run(
        "docker exec master_spark /spark/bin/spark-submit "
        "--jars /root/download/nebula-algo.jar "
        "/root/run/scan_reader.py",
        shell=True,
        check=True,
        capture_output=True,
    )

    assert result.returncode == 0, f"ERROR during run scan_reader.py: {result}"
    print(f"Scan reader result:\n{result.stdout.decode('utf-8')}")


def test_scan_reader_spark_engine_and_run_pagerank():
    """
    Just call:
    /spark/bin/spark-submit \
        --jars /root/download/nebula-algo.jar \
        /root/run/algo.py
    in container: master_spark
    assert the return value is 0
    """

    result = subprocess.run(
        "docker exec master_spark /spark/bin/spark-submit "
        "--jars /root/download/nebula-algo.jar "
        "/root/run/algo.py",
        shell=True,
        check=True,
        capture_output=True,
    )

    assert result.returncode == 0, f"ERROR during run algo.py: {result}"
    print(f"Scan reader result:\n{result.stdout.decode('utf-8')}")


def test_query_reader_spark_engine():
    """
    Just call:
    /spark/bin/spark-submit \
        --jars /root/download/nebula-algo.jar \
        /root/run/query_reader.py
    in container: master_spark
    assert the return value is 0
    """

    result = subprocess.run(
        "docker exec master_spark /spark/bin/spark-submit "
        "--jars /root/download/nebula-algo.jar "
        "/root/run/query_reader.py",
        shell=True,
        check=True,
        capture_output=True,
    )

    assert result.returncode == 0, f"ERROR during run query_reader.py: {result}"
    print(f"Query reader result:\n{result.stdout.decode('utf-8')}")


def test_label_propagation_spark_engine_writer():
    """
    Just call:
    /spark/bin/spark-submit \
        --jars /root/download/nebula-algo.jar \
        /root/run/writer.py
    in container: master_spark
    assert the return value is 0
    Then query NebulaGraph on tag: label_propagation
    assert the result is correct
    """

    result = subprocess.run(
        "docker exec master_spark /spark/bin/spark-submit "
        "--jars /root/download/nebula-algo.jar "
        "/root/run/writer.py",
        shell=True,
        check=True,
        capture_output=True,
    )

    assert result.returncode == 0, f"ERROR during run writer.py: {result}"
    print(f"Label propagation result:\n{result.stdout.decode('utf-8')}")

    nebula_config = Config()
    connection_pool = ConnectionPool()
    connection_pool.init([("127.0.0.1", 39669)], nebula_config)

    with connection_pool.session_context("root", "nebula") as session:
        session.execute("USE basketballplayer")
        result = session.execute(
            "MATCH (v:player) RETURN v.label_propagation.cluster_id LIMIT 1"
        )
        print(result)

    assert result.is_succeeded(), f"ERROR during query NebulaGraph: {result}"
    assert (
        not result.is_empty()
    ), f"label_propagation not written to NebulaGraph result: {result}"
    assert (
        result.column_values("v.label_propagation.cluster_id")[0]
        .cast()
        .startswith("player")
    ), f"label_propagation value is not correct result: {result}"
    print(f"Label propagation result:\n{result}")

    # edge writer
    with connection_pool.session_context("root", "nebula") as session:
        session.execute("USE basketballplayer")
        result = session.execute(
            "MATCH ()-[e:jaccard_similarity]->() RETURN e.similarity LIMIT 1"
        )
        print(result)
    connection_pool.close()
    assert result.is_succeeded(), f"ERROR during query NebulaGraph: {result}"
    assert (
        0 < result.column_values("e.similarity")[0].cast() < 1
    ), f"jaccard_similarity value is not correct result: {result}"



from unittest.mock import Mock

import pytest

from ng_ai.config import NebulaGraphConfig
from ng_ai.engines import SparkEngine
from ng_ai.nebula_data import NebulaDataFrameObject
from ng_ai.nebula_reader import NebulaReader, NebulaReaderWithSpark


@pytest.fixture
def nebula_reader_with_spark():
    config = NebulaGraphConfig()
    return NebulaReaderWithSpark(config=config)


def test_nebula_reader_init():
    config = NebulaGraphConfig()
    reader = NebulaReader(engine="spark", config=config)

    assert isinstance(reader, NebulaReader)
    assert isinstance(reader.reader, NebulaReaderWithSpark)


def test_scan(nebula_reader_with_spark):
    kwargs = {"edge": "follow", "props": "degree"}

    nebula_reader_with_spark.scan(**kwargs)

    assert nebula_reader_with_spark.raw_df_reader is not None
    assert isinstance(nebula_reader_with_spark.engine, SparkEngine)

    spark = nebula_reader_with_spark.engine.spark
    assert spark is not None


def test_query(nebula_reader_with_spark):
    kwargs = {
        "query": "MATCH ()-[e:follow]->() RETURN e LIMIT 100000",
        "edge": "follow",
        "props": "degree",
    }

    nebula_reader_with_spark.query(**kwargs)

    assert nebula_reader_with_spark.raw_df_reader is not None
    assert isinstance(nebula_reader_with_spark.engine, SparkEngine)

    spark = nebula_reader_with_spark.engine.spark
    assert spark is not None


def test_read(nebula_reader_with_spark):
    data = [[1, "foo"], [2, "bar"]]
    mock_raw_df_reader = Mock()
    mock_raw_df_reader.load.return_value = data
    nebula_reader_with_spark.raw_df_reader = mock_raw_df_reader

    result = nebula_reader_with_spark.read()

    assert mock_raw_df_reader.load.called
    assert isinstance(result, NebulaDataFrameObject)
    assert nebula_reader_with_spark.raw_df == data



from unittest.mock import Mock, patch

import pytest
from pyspark.sql import SparkSession

from ng_ai.config import NebulaGraphConfig
from ng_ai.nebula_writer import NebulaWriter, NebulaWriterWithSpark


@pytest.fixture(scope="session")
def spark():
    spark = (
        SparkSession.builder.master("local[1]")
        .config("spark.port.maxRetries", "1000")
        .config("spark.sql.shuffle.partitions", "1")
        .getOrCreate()
    )
    return spark


@pytest.fixture(scope="session")
def spark_df(spark):
    """
    +---------+---------+-----+------+
    |   _srcId|   _dstId|_rank|degree|
    +---------+---------+-----+------+
    |player105|player100|    0|    70|
    |player105|player104|    0|    83|
    +---------+---------+-----+------+
    """
    data = [
        ("player105", "player100", 0, 70),
        ("player105", "player104", 0, 83),
    ]
    df = spark.createDataFrame(data, ["_srcId", "_dstId", "_rank", "degree"])
    return df


# TBD, need to test writer with data: NebulaDataFrameObject
# @pytest.fixture
# def nebula_writer_with_spark_nebula_df():
#     config = NebulaGraphConfig()
#     spark_engine = SparkEngine(config=config)
#     data = NebulaDataFrameObject(engine=spark_engine, data=spark_df)
#     return NebulaWriter(data, "nebulagraph_vertex", config=config, engine="spark")


@pytest.fixture
def nebula_writer_with_spark_jdf(spark_df):
    config = NebulaGraphConfig()
    data = spark_df._jdf
    return NebulaWriter(data, "nebulagraph_vertex", config=config, engine="spark")


def test_nebula_writer_init(spark_df):
    config = NebulaGraphConfig()
    data = spark_df._jdf
    writer = NebulaWriter(data, "nebulagraph_vertex", config=config, engine="spark")

    assert isinstance(writer, NebulaWriter)
    assert isinstance(writer.writer, NebulaWriterWithSpark)
    assert writer.jdf is not None


@patch("ng_ai.nebula_writer.NebulaWriterWithSpark._get_raw_df_writer_with_nebula")
def test_set_options_with_nebula(
    mock_get_raw_df_writer, nebula_writer_with_spark_jdf
):
    mock_raw_df_writer = Mock()
    mock_get_raw_df_writer.return_value = mock_raw_df_writer

    properties = {"lpa": "cluster_id"}

    kwargs = {
        "space": "basketballplayer",
        "type": "vertex",
        "tag": "person",
        "write_mode": "insert",
        "batch": 128,
        "vid_policy": "",
        "vid_field": "id",
        "properties": properties,
    }
    nebula_writer_with_spark_jdf.set_options(**kwargs)

    assert mock_get_raw_df_writer.called
    assert mock_raw_df_writer.option.called
    assert mock_raw_df_writer.option.call_count == 11


@patch("ng_ai.nebula_writer.NebulaWriterWithSpark._get_raw_df_writer_with_nebula")
def test_write(mock_get_raw_df_writer, nebula_writer_with_spark_jdf):
    mock_raw_df_writer = Mock()
    mock_get_raw_df_writer.return_value = mock_raw_df_writer

    # set up options for writing
    properties = {"lpa": "cluster_id"}

    kwargs = {
        "space": "basketballplayer",
        "type": "vertex",
        "tag": "person",
        "write_mode": "insert",
        "batch": 128,
        "vid_policy": "",
        "vid_field": "id",
        "properties": properties,
    }

    nebula_writer_with_spark_jdf.set_options(**kwargs)

    with patch.object(mock_raw_df_writer, "save") as mock_save:
        nebula_writer_with_spark_jdf.write()
        assert mock_save.called



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.

from __future__ import annotations

from ng_ai.nebula_data import NebulaDataFrameObject as NebulaDataFrameObjectImpl
from ng_ai.nebula_data import NebulaGraphObject as NebulaGraphObjectImpl


def algo(func):
    func.is_algo = True
    return func


class NebulaAlgorithm:
    def __init__(self, obj: NebulaGraphObjectImpl or NebulaDataFrameObjectImpl):
        if isinstance(obj, NebulaGraphObjectImpl):
            self.algorithm = NebulaGraphAlgorithm(obj)
        elif isinstance(obj, NebulaDataFrameObjectImpl):
            self.algorithm = NebulaDataFrameAlgorithm(obj)
        else:
            raise ValueError(
                f"Unsupported object type: {type(obj)}",
                "NebulaAlgorithm only supports NebulaGraphObject and"
                "NebulaDataFrameObject",
            )
        self.obj = obj
        self.engine = obj.engine

    def __getattr__(self, name):
        return getattr(self.algorithm, name)


class NebulaDataFrameAlgorithm:
    """
    Spark Dataframe to run algorithm
    """

    def __init__(self, ndf_obj: NebulaDataFrameObjectImpl):
        self.ndf_obj = ndf_obj
        self.algorithms = []

    def register_algo(self, func):
        self.algorithms.append(func.__name__)

    def get_all_algo(self):
        if not self.algorithms:
            for name, func in NebulaDataFrameAlgorithm.__dict__.items():
                if hasattr(func, "is_algo"):
                    self.register_algo(func)
        return self.algorithms

    def check_engine(self):
        """
        Check if the engine is supported.
        For netowrkx, we need to convert the NebulaDataFrameObject
            to NebulaGraphObject
        For spark, we can directly use the NebulaDataFrameObject
        """
        if self.ndf_obj.engine.type == "nebula":
            raise Exception(
                "For NebulaDataFrameObject in networkx engine,"
                "Plz transform it to NebulaGraphObject to run algorithm",
                "For example: g = nebula_df.to_graph; g.algo.pagerank()",
            )

    def get_spark_engine_context(self, config_class: str, lib_class: str):
        """
        Get the engine context
        """
        self.check_engine()
        engine = self.ndf_obj.engine
        spark = engine.spark
        jspark = engine.jspark
        engine.import_algo_config_class(config_class)
        engine.import_algo_lib_class(lib_class)
        return engine, spark, jspark, engine.encode_vid

    def get_spark_dataframe(self):
        """
        Check if df is a pyspark.sql.dataframe.DataFrameme, and return it
        """
        df = self.ndf_obj.data
        from pyspark.sql.dataframe import DataFrame as pyspark_sql_df

        if not isinstance(df, pyspark_sql_df):
            raise Exception(
                "The NebulaDataFrameObject is not a spark dataframe",
                f"Got type(df): {type(df)}",
            )
        return df

    @algo
    def pagerank(
        self, reset_prob: float = 0.15, max_iter: int = 10, weighted: bool = False
    ):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "PRConfig", "PageRankAlgo"
        )
        df = self.get_spark_dataframe()
        config = spark._jvm.PRConfig(max_iter, reset_prob, encode_vid)
        result = spark._jvm.PageRankAlgo.apply(jspark, df._jdf, config, weighted)

        return result

    @algo
    def connected_components(self, max_iter: int = 10, weighted: bool = False):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "CcConfig", "ConnectedComponentsAlgo"
        )
        df = self.get_spark_dataframe()
        config = spark._jvm.CcConfig(max_iter, encode_vid)
        result = spark._jvm.ConnectedComponentsAlgo.apply(
            jspark, df._jdf, config, weighted
        )

        return result

    @algo
    def label_propagation(self, max_iter: int = 10, weighted: bool = False):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "LPAConfig", "LabelPropagationAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.LPAConfig(max_iter, encode_vid)
        result = spark._jvm.LabelPropagationAlgo.apply(
            jspark, df._jdf, config, weighted
        )

        return result

    @algo
    def louvain(self, max_iter: int = 20, internalIter: int = 10, tol: float = 0.5):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "LouvainConfig", "LouvainAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.LouvainConfig(max_iter, internalIter, tol, encode_vid)
        result = spark._jvm.LouvainAlgo.apply(jspark, df._jdf, config, False)

        return result

    @algo
    def k_core(self, max_iter: int = 10, degree: int = 2):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "KCoreConfig", "KCoreAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.KCoreConfig(max_iter, degree, encode_vid)

        result = spark._jvm.KCoreAlgo.apply(jspark, df._jdf, config)

        return result

    # def shortest_path(self, landmarks: list, weighted: bool = False):
    #     engine, spark, jspark, encode_vid = self.get_spark_engine_context(
    #         "ShortestPathConfig", "ShortestPathAlgo"
    #     )
    #     # TBD: ShortestPathAlgo is not yet encodeID compatible
    #     df = self.get_spark_dataframe()

    #     config = spark._jvm.ShortestPathConfig(landmarks, encode_vid)
    #     result = spark._jvm.ShortestPathAlgo.apply(
    #         jspark, df._jdf, config, weighted)

    #     return result

    @algo
    def degree_statics(self):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "DegreeStaticConfig", "DegreeStaticAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.DegreeStaticConfig(encode_vid)
        result = spark._jvm.DegreeStaticAlgo.apply(jspark, df._jdf, config)

        return result

    @algo
    def betweenness_centrality(self, max_iter: int = 10, weighted: bool = False):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "BetweennessConfig", "BetweennessCentralityAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.BetweennessConfig(max_iter, encode_vid)
        result = spark._jvm.BetweennessCentralityAlgo.apply(
            jspark, df._jdf, config, weighted
        )

        return result

    @algo
    def clustering_coefficient(self, type: str = "local"):
        # type could be either "local" or "global"
        assert type.lower() in ["local", "global"], (
            "type should be either local or global"
            f"in clustering_coefficient algo. Got type: {type}"
        )
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "CoefficientConfig", "ClusteringCoefficientAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.CoefficientConfig(type, encode_vid)
        result = spark._jvm.ClusteringCoefficientAlgo.apply(jspark, df._jdf, config)

        return result

    @algo
    def bfs(self, max_depth: int = 10, root: int = 1):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "BfsConfig", "BfsAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.BfsConfig(max_depth, root, encode_vid)
        result = spark._jvm.BfsAlgo.apply(jspark, df._jdf, config)

        return result

    # dfs is not yet supported, need to revisit upstream nebula-algorithm
    @algo
    def dfs(self, max_depth: int = 10, root: int = 1):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "DfsConfig", "DfsAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.DfsConfig(max_depth, root, encode_vid)
        result = spark._jvm.DfsAlgo.apply(jspark, df._jdf, config)

        return result

    @algo
    def hanp(
        self,
        hop_attenuation: float = 0.5,
        max_iter: int = 10,
        preference: float = 1.0,
        weighted: bool = False,
        preferences=None,
    ):
        """
        Hop Attenuation & Node Preference
        """
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "HanpConfig", "HanpAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.HanpConfig(
            hop_attenuation, max_iter, preference, encode_vid
        )
        result = spark._jvm.HanpAlgo.apply(
            jspark, df._jdf, config, weighted, preferences
        )

        return result

    # @algo
    # def node2vec(
    #     self,
    #     max_iter: int = 10,
    #     lr: float = 0.025,
    #     data_num_partition: int = 10,
    #     model_num_partition: int = 10,
    #     dim: int = 10,
    #     window: int = 3,
    #     walk_length: int = 5,
    #     num_walks: int = 3,
    #     p: float = 1.0,
    #     q: float = 1.0,
    #     directed: bool = False,
    #     degree: int = 30,
    #     emb_separator: str = ",",
    #     model_path: str = "hdfs://127.0.0.1:9000/model",
    #     weighted: bool = False,
    # ):
    #     engine, spark, jspark, encode_vid = self.get_spark_engine_context(
    #         "Node2vecConfig", "Node2VecAlgo"
    #     )
    #     # TBD: Node2VecAlgo is not yet encodeID compatible
    #     df = self.get_spark_dataframe()
    #     config = spark._jvm.Node2vecConfig(
    #         max_iter,
    #         lr,
    #         data_num_partition,
    #         model_num_partition,
    #         dim,
    #         window,
    #         walk_length,
    #         num_walks,
    #         p,
    #         q,
    #         directed,
    #         degree,
    #         emb_separator,
    #         model_path,
    #         encode_vid,
    #     )
    #     result = spark._jvm.Node2VecAlgo.apply(jspark, df._jdf, config, weighted)

    #     return result

    @algo
    def jaccard(self, tol: float = 1.0):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "JaccardConfig", "JaccardAlgo"
        )
        df = self.get_spark_dataframe()

        config = spark._jvm.JaccardConfig(tol, encode_vid)
        result = spark._jvm.JaccardAlgo.apply(jspark, df._jdf, config)

        return result

    @algo
    def strong_connected_components(
        self, max_iter: int = 10, weighted: bool = False
    ):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "CcConfig", "StronglyConnectedComponentsAlgo"
        )
        df = self.get_spark_dataframe()
        config = spark._jvm.CcConfig(max_iter, encode_vid)
        result = spark._jvm.StronglyConnectedComponentsAlgo.apply(
            jspark, df._jdf, config, weighted
        )

        return result

    @algo
    def triangle_count(self):
        engine, spark, jspark, encode_vid = self.get_spark_engine_context(
            "TriangleConfig", "TriangleCountAlgo"
        )
        df = self.get_spark_dataframe()
        config = spark._jvm.TriangleConfig(encode_vid)
        result = spark._jvm.TriangleCountAlgo.apply(jspark, df._jdf, config)

        return result

    # @algo
    # def closeness_centrality(self, weighted: bool = False):
    #     # TBD: ClosenessAlgo is not yet encodeID compatible
    #     engine, spark, jspark, encode_vid = self.get_spark_engine_context(
    #         "ClosenessConfig", "ClosenessAlgo"
    #     )
    #     df = self.get_spark_dataframe()
    #     config = spark._jvm.ClosenessConfig(weighted, encode_vid)
    #     result = spark._jvm.ClosenessAlgo.apply(
    #         jspark, df._jdf, config, False)

    #     return result


class NebulaGraphAlgorithm:
    """
    Networkx to run algorithm
    """

    def __init__(self, ng_obj: NebulaGraphObjectImpl):
        self.ngraph = ng_obj
        self.algorithms = []
        self.engine = ng_obj.engine

    def register_algo(self, func):
        self.algorithms.append(func.__name__)

    def get_all_algo(self):
        if not self.algorithms:
            for name, func in NebulaGraphAlgorithm.__dict__.items():
                if hasattr(func, "is_algo"):
                    self.register_algo(func)
        return self.algorithms

    def check_engine(self):
        """
        Check if the engine is supported.
        For netowrkx, we can directly call .algo.pagerank()
        For spark, we need to convert the NebulaGraphObject
            to NebulaDataFrameObject
        """
        if self.engine.type == "spark":
            raise Exception(
                "For NebulaGraphObject in spark engine,"
                "Plz transform it to NebulaDataFrameObject to run algorithm",
                "For example: df = nebula_graph.to_df; df.algo.pagerank()",
            )
        if self.engine.type == "nebula":
            return True
        else:
            raise Exception("Unsupported engine type")

    @algo
    def pagerank(self, reset_prob=0.15, max_iter=10, **kwargs):
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        weight = kwargs.get("weight", None)
        assert type(weight) in [str, type(None)], "weight must be str or None"
        assert type(reset_prob) == float, "reset_prob must be float"
        assert type(max_iter) == int, "max_iter must be int"
        tol = kwargs.get("tol", 1e-06)
        assert type(tol) == float, "tol must be float"

        return self.engine.nx.pagerank(
            g, alpha=1 - reset_prob, max_iter=max_iter, tol=tol, weight=weight
        )

    @algo
    def louvain(self, weight: str = None, resolution: float = 1.0):
        """
        doc: https://github.com/taynaud/python-louvain
        """
        weight = weight if weight else ""
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        ug = g.to_undirected()
        return self.engine.nx_community_louvain.best_partition(
            ug, weight=weight, resolution=resolution
        )

    @algo
    def label_propagation(self, **kwargs):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /community.html
        """

        self.check_engine()
        g = self.ngraph.get_nx_graph()
        ug = g.to_undirected()
        return self.engine.nx.algorithms.community.label_propagation_communities(
            ug, **kwargs
        )

    @algo
    def k_core(self, k: int = 2):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /generated/networkx.algorithms.core.k_core.html
        return: networkx.classes.digraph.DiGraph
        """
        # TBD, k_core requires single graph
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        # MultiDiGraph to DiGraph
        single_g = self.engine.nx.DiGraph()
        for u, v in g.edges():
            single_g.add_edge(u, v)
        return self.engine.nx.k_core(single_g, k=k)

    @algo
    def k_truss(self, k: int = 2):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /generated/networkx.algorithms.core.k_truss.html
        return: networkx.classes.graph.Graph
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        # MultiDiGraph to Graph
        single_ug = self.engine.nx.Graph()
        for u, v in g.edges():
            single_ug.add_edge(u, v)
        return self.engine.nx.k_truss(single_ug, k=k)

    @algo
    def k_clique_communities(self, k: int = 2, **kwargs):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /community.html
        return: Yields sets of nodes, one for each k-clique community.
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        ug = g.to_undirected()
        return self.engine.nx.algorithms.community.k_clique_communities(
            ug, k=k, **kwargs
        )

    @algo
    def degree_statics(self):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/classes
        /generated/networkx.MultiDiGraph.degree.html
        return:
          List[Tuple[str, int, int, int]],
          (node_id, degree, in_degree, out_degree)
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        degrees = dict(g.degree())
        in_degrees = dict(g.in_degree())
        out_degrees = dict(g.out_degree())

        result = []
        for node_id in g.nodes():
            degree = degrees[node_id]
            in_degree = in_degrees[node_id]
            out_degree = out_degrees[node_id]
            result.append((node_id, degree, in_degree, out_degree))
        return result

    @algo
    def betweenness_centrality(
        self, k: int = None, normalized: bool = True, weight: str = None, **kwargs
    ):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /generated/networkx.algorithms.centrality.betweenness_centrality.html
        return: Dictionary of nodes with betweenness centrality as the value.
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        # MultiDiGraph to DiGraph
        single_g = self.engine.nx.DiGraph()
        for u, v, data in g.edges(data=True):
            if weight is not None and weight in data:
                single_g.add_edge(u, v, weight=data[weight])
            else:
                single_g.add_edge(u, v)
        return self.engine.nx.betweenness_centrality(
            single_g, k=k, normalized=normalized, weight=weight, **kwargs
        )

    @algo
    def clustering_coefficient(self, weight: str = None, **kwargs):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /generated/networkx.algorithms.cluster.clustering.html
        return: Dictionary of nodes with clustering coefficient as the value.
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        # MultiDiGraph to DiGraph
        single_g = self.engine.nx.DiGraph()
        for u, v, data in g.edges(data=True):
            if weight is not None and weight in data:
                single_g.add_edge(u, v, weight=data[weight])
            else:
                single_g.add_edge(u, v)
        return self.engine.nx.clustering(single_g, weight=weight, **kwargs)

    @algo
    def bfs(self, root=None, max_depth: int = 10, reverse: bool = False, **kwargs):
        """
        doc:
        https://networkx.org/documentation/networkx-2.6.2/reference/algorithms/
        generated/networkx.algorithms.traversal.breadth_first_search.bfs_edges.html
        root: The node at which to start the search,
          defaults to the first node in the graph.
        reverse: If True, perform a reverse breadth-first-search.
        return: Yields edges in a breadth-first-search starting at source.
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        if root is None:
            root = next(iter(g.nodes()))
        return self.engine.nx.bfs_edges(
            g, source=root, depth_limit=max_depth, reverse=reverse, **kwargs
        )

    @algo
    def dfs(self, root=None, max_depth: int = 10, **kwargs):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /generated/networkx.algorithms.traversal.depth_first_search.dfs_edges.html
        root: The node at which to start the search,
          defaults to the first node in the graph.
        return: Yields edges in a depth-first-search starting at source.
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        if root is None:
            root = next(iter(g.nodes()))
        return self.engine.nx.dfs_edges(
            g, source=root, depth_limit=max_depth, **kwargs
        )

    @algo
    def node2vec(
        self,
        dimensions: int = 128,
        walk_length: int = 80,
        num_walks: int = 10,
        workers: int = 1,
        fit_args: dict = {},
        **kwargs,
    ):
        """
        doc: https://github.com/eliorc/node2vec
        dimensions: int, optional (default = 128), Dimensionality of the
          word vectors.
        walk_length: int, optional (default = 80), Length of walk per source.
          Default value is 80.
        num_walks: int, optional (default = 10), Number of walks per source.
          Default value is 10.
        workers: int, optional (default = 1), Number of parallel workers.
          Default is 1.
        fit_args: dict, optional (default = {}), Arguments for
          gensim.models.Word2Vec.fit()
        return: gensim.models.keyedvectors.Word2VecKeyedVectorsmodel
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        node2vec = self.engine.nx_node2vec(
            graph=g,
            dimensions=dimensions,
            walk_length=walk_length,
            num_walks=num_walks,
            workers=workers,
            **kwargs,
        )
        model = node2vec.fit(**fit_args)
        return model

    @algo
    def jaccard(self, ebunch: list = None, **kwargs):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /generated/networkx.algorithms.link_prediction.jaccard_coefficient.html
        ebunch: iterable of node pairs, optional (default = None),
          If provided, only return the Jaccard coefficient for the specified pairs.
            example: [('A', 'B'), ('A', 'C')]
        return: Yields tuples of (u, v, p) where u and v are nodes
          and p is the Jaccard coefficient of the neighbors of u and v.
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        # MultiDiGraph to Graph
        single_ug = self.engine.nx.Graph()
        for u, v in g.edges():
            single_ug.add_edge(u, v)
        return self.engine.nx.jaccard_coefficient(
            single_ug, ebunch=ebunch, **kwargs
        )

    @algo
    def connected_components(self, **kwargs):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /generated/networkx.algorithms.components.connected_components.html
        return: A generator of sets of nodes, one for each connected component
          in the graph.
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        # MultiDiGraph to MultiGraph
        ug = g.to_undirected()
        return self.engine.nx.connected_components(ug, **kwargs)

    @algo
    def weakly_connected_components(self, **kwargs):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /generated/networkx.algorithms.components.weakly_connected_components.html
        return: A generator of sets of nodes, one for each weakly connected
          component in the graph.
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        return self.engine.nx.weakly_connected_components(g, **kwargs)

    @algo
    def strongly_connected_components(self, **kwargs):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /generated/networkx.algorithms.components.strongly_connected_components.html
        return: A generator of sets of nodes, one for each strongly connected
          component in the graph.
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        return self.engine.nx.strongly_connected_components(g, **kwargs)

    @algo
    def triangle_count(self, **kwargs):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /generated/networkx.algorithms.cluster.triangles.html
        return: A dictionary keyed by node to the number of triangles that include
          that node as a vertex.
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        # MultiDiGraph to Graph
        single_ug = self.engine.nx.Graph()
        for u, v in g.edges():
            single_ug.add_edge(u, v)
        return self.engine.nx.triangles(single_ug, **kwargs)

    @algo
    def closeness_centrality(
        self, u=None, distance=None, wf_improved=True, **kwargs
    ):
        """
        doc: https://networkx.org/documentation/networkx-2.6.2/reference/algorithms
        /generated/networkx.algorithms.centrality.closeness_centrality.html
        u: node, optional (default = None), If specified, return only the value
            for the node u.
        distance: edge attribute key, optional (default = None), Use the specified
            edge attribute as the edge distance in shortest path calculations.
        wf_improved: bool, optional (default = True), If True, use the improved
            algorithm of Freeman and Bader which computes the closeness centrality
            using the number of reachable nodes instead of the number of nodes in
            the graph.
        return: A dictionary keyed by node to the closeness centrality of that
            node.
        """
        self.check_engine()
        g = self.ngraph.get_nx_graph()
        return self.engine.nx.closeness_centrality(
            g, u=u, distance=distance, wf_improved=wf_improved, **kwargs
        )



import os
import subprocess
from time import sleep

import pytest


@pytest.fixture(scope="session", autouse=True)
def prepare_data():
    print("Setup data...")
    subprocess.run(
        "docker run --rm --network setup_ngai-net vesoft/nebula-console:v3"
        ' -addr graphd -port 9669 -u root -p nebula -e " '
        'DROP SPACE basketballplayer;" ',
        shell=True,
        check=True,
    )
    sleep(4)
    result = subprocess.run(
        "docker run --rm --network setup_ngai-net vesoft/nebula-console:v3"
        ' -addr graphd -port 9669 -u root -p nebula -e " '
        ':play basketballplayer" ',
        shell=True,
        check=True,
        capture_output=True,
        timeout=300,
    )
    print(f":play basketballplayer: {result}")

    subprocess.run(
        "docker run --rm --network setup_ngai-net vesoft/nebula-console:v3"
        ' -addr graphd -port 9669 -u root -p nebula -e " '
        'USE basketballplayer; SUBMIT JOB STATS" ',
        shell=True,
        check=True,
        capture_output=True,
    )

    result = subprocess.run(
        "docker run --rm --network setup_ngai-net vesoft/nebula-console:v3"
        ' -addr graphd -port 9669 -u root -p nebula -e " '
        "USE basketballplayer; "
        "CREATE TAG IF NOT EXISTS label_propagation "
        '(cluster_id string NOT NULL);" ',
        shell=True,
        check=True,
        capture_output=True,
    )
    sleep(4)
    assert (
        b"ERROR" not in result.stdout
    ), f"ERROR during create tag: {result.stdout.decode('utf-8')}"

    result = subprocess.run(
        "docker run --rm --network setup_ngai-net vesoft/nebula-console:v3"
        ' -addr graphd -port 9669 -u root -p nebula -e " '
        "USE basketballplayer; "
        "CREATE EDGE IF NOT EXISTS jaccard_similarity "
        '(similarity double);" ',
        shell=True,
        check=True,
        capture_output=True,
    )
    sleep(4)
    assert (
        b"ERROR" not in result.stdout
    ), f"ERROR during create edge: {result.stdout.decode('utf-8')}"

    result = subprocess.run(
        "docker run --rm --network setup_ngai-net vesoft/nebula-console:v3"
        ' -addr graphd -port 9669 -u root -p nebula -e " '
        "USE basketballplayer; "
        'SHOW STATS;" ',
        shell=True,
        check=True,
        capture_output=True,
    )
    print(f"Show stats:\n{result.stdout.decode('utf-8')}")

    os.system("mkdir -p tests/integration/setup/run")
    os.system("rm -rf tests/integration/setup/run/*")

    os.system(
        "cp -r tests/integration/spark_engine_cases/* tests/integration/setup/run/"
    )

    # create louvain schema for networkx engine writer test
    # RUN:
    # CREATE TAG IF NOT EXISTS louvain (
    #     cluster_id string NOT NULL
    # );

    result = subprocess.run(
        "docker run --rm --network setup_ngai-net vesoft/nebula-console:v3"
        ' -addr graphd -port 9669 -u root -p nebula -e " '
        "USE basketballplayer; "
        "CREATE TAG IF NOT EXISTS louvain "
        '(cluster_id string NOT NULL);" ',
        shell=True,
        check=True,
        capture_output=True,
    )
    sleep(10)
    # check the schema existence with DESC TAG louvain:
    result = subprocess.run(
        "docker run --rm --network setup_ngai-net vesoft/nebula-console:v3"
        ' -addr graphd -port 9669 -u root -p nebula -e " '
        "USE basketballplayer; "
        'DESC TAG louvain;" ',
        shell=True,
        check=True,
        capture_output=True,
    )
    print(f"DESC TAG louvain:\n{result.stdout.decode('utf-8')}")
    assert (
        b"ERROR" not in result.stdout
    ), f"ERROR during create tag: {result.stdout.decode('utf-8')}"

    # check the schema existence with DESC EDGE jaccard_similarity:
    result = subprocess.run(
        "docker run --rm --network setup_ngai-net vesoft/nebula-console:v3"
        ' -addr graphd -port 9669 -u root -p nebula -e " '
        "USE basketballplayer; "
        'DESC EDGE jaccard_similarity;" ',
        shell=True,
        check=True,
        capture_output=True,
    )
    print(f"DESC EDGE jaccard_similarity:\n{result.stdout.decode('utf-8')}")
    assert (
        b"ERROR" not in result.stdout
    ), f"ERROR during create edge: {result.stdout.decode('utf-8')}"

    print("Setup data done...")



from ng_ai import NebulaGraphConfig, NebulaReader

# read data with spark engine, query mode
config_dict = {
    "graphd_hosts": "graphd:9669",
    "metad_hosts": "metad0:9559",
    "user": "root",
    "password": "nebula",
    "space": "basketballplayer",
}
config = NebulaGraphConfig(**config_dict)
reader = NebulaReader(engine="spark", config=config)
query = """
    MATCH ()-[e:follow]->()
    RETURN e LIMIT 100000
"""
reader.query(query=query, edge="follow", props="degree")
df = reader.read()

"""
/spark/bin/spark-submit \
    --jars /root/download/nebula-algo.jar \
    /root/run/query_reader.py
"""
